#include <sstream>
#include <cctype>
#include <cstdio>
#include <string>
#include <vector>
#include <list>
#include <map>

#undef SUPPORT_INSERTS

using namespace std;

#ifdef SUPPORT_INSERTS
static bool nodata = true;
#endif

static const bool startwith(const char *s2, const string &s, unsigned ind=0)
{
    unsigned a;
    for(a=ind; s2 && *s2; ++a)
    {
        if(a >= s.size())return false;
        if(toupper(s[a]) != *s2++)
            return false;
    }
    return true;
}

static const string ucase(const string &s)
{
    string t;
    unsigned a, b=s.size();
    for(a=0; a<b; ++a)t += toupper(s[a]);
    return t;
}

static const string GetStr(const string &s, unsigned &a)
{
    string t;
    const unsigned b = s.size();
    if(a < b && s[a] == '`')
    {
        for(++a; a<b && s[a] != '`'; ) t += s[a++];
        if(a < b && s[a] == '`') ++a;
    }
    else
    {
        while(a < b && (isalnum(s[a]) || s[a]=='_'))
            t += s[a++];
    }
    return t;
}
static const bool ischar(char c)
{
    return c=='_' || isalpha(c);
}

static const string sqlfix(const string &s)
{
    bool is_int_without_zerolead = true;
    if(s.empty()) is_int_without_zerolead = false;
    for(unsigned zeroes=0, nonzeroes=0, a=0; a<s.size(); ++a)
    {
        if((s[a]<'0' || s[a]>'9')
        || (s[a]=='0' && zeroes>0 && nonzeroes==0))
        {
            is_int_without_zerolead = false;
            break;
        }
        if(!a && s[a]=='0') ++zeroes; else ++nonzeroes;
    }
    if(is_int_without_zerolead) return s;

    string t;
    unsigned a, b=s.size();
    for(a=0; a<b; ++a)
    {
        if(s[a]=='\\' || s[a]=='\'')t += '\\';
        t += s[a];
    }
    return "'" + t + "'";
}

class fieldtype
{
    string type;
    unsigned width, prec;
    bool autoincrement;
    bool zerofill;
    bool unsig;
    bool null;
    string deftext;
    string content;
    bool binary, unique, fulltext;
    void RebuildContent()
    {
        content = type;
        if(width)
        {
            char Buf[64];
            content += '(';
            sprintf(Buf, "%u", width); content += Buf;
            if(prec) { sprintf(Buf, ",%u", prec); content += Buf; }
            content += ')';
        }
        content += ' ';
        if(zerofill)content += "ZEROFILL ";
        if(unsig)content += "UNSIGNED ";
        if(binary)content += "BINARY ";
        if(unique)content += "UNIQUE ";
        if(fulltext)content += "FULLTEXT ";
        if(null)content += "NULL "; else { content += "NOT NULL "; forcedefault(); }
        if(autoincrement)
        {
            content += "AUTO_INCREMENT ";
        }
    }
public:
    inline const bool operator!= (const fieldtype &s) const { return !operator==(s); }
    const bool operator== (const fieldtype &s) const
    {
        return s.width==width
           &&  s.prec==prec
           &&  s.autoincrement==autoincrement
           &&  s.zerofill==zerofill
           &&  s.unsig==unsig
           &&  s.null==null
           &&  s.type==type
           &&  s.binary==binary;
    }
    const char *c_str() const { return content.c_str(); }
    const string& defval() const { return deftext; }
    void forcedefault()
    {
        if(!deftext.empty()) return;
        
        if(type=="INT" || type=="TINYINT" || type=="SMALLINT" || type=="BIGINT")
        {
            deftext = sqlfix("0");
        }
        else if(type=="DECIMAL" || type=="DOUBLE" || type=="FLOAT")
        {
            deftext = "0";
            if(prec) deftext += '.';
            for(unsigned a=0; a<prec; ++a) deftext += '0';
            deftext = sqlfix(deftext);
        }
        else
            deftext = sqlfix("");
    }
    const string &gettype() const { return type; }
    const unsigned getwidth() const { return width; }
    void settype(const string &t) { type=t; RebuildContent(); }

    explicit fieldtype(const string &s) : width(0), prec(0), autoincrement(false),
                                          zerofill(false), unsig(false), null(true),
                                          content(s), binary(false), unique(false),
                                          fulltext(false)
    {
        unsigned b=s.size(), c=0;
        for(; c<b && isspace(s[c]); ++c);
        type = ucase(GetStr(s, c));
        if(type.empty())
        {
            throw("Can't get type from '" + s + "'");
        }
        if(type=="INTEGER")type="INT";
        else if(type=="REAL")type="DOUBLE";
        else if(type=="NUMERIC")type="DECIMAL";

        if(type=="INT")width=11;
        else if(type=="SMALLINT")width=6;
        else if(type=="TINYINT")width=4;
        else if(type=="MEDIUMINT")width=9;
        else if(type=="BIGINT")width=20;
        else if(type=="DOUBLE")width=16, prec=4;
        else if(type=="FLOAT")width=10, prec=2;
        else if(type=="DECIMAL"){}

        else if(type=="CHAR")width=1;
        else if(type=="VARCHAR"){}
        else if(type=="TEXT"){}
        else if(type=="BLOB"){}
        else if(type=="LONGBLOB"){}
        else
        {
            fprintf(stderr, "Warning: Unknown type '%s'\n", type.c_str());
        }
        for(; c<b && isspace(s[c]); ++c);
        if(c<b)
        {
            if(s[c]=='(')
            {
                for(++c; c<b && isspace(s[c]); ++c);
                for(width=0; c<b && isdigit(s[c]); ++c)width=width*10+s[c]-'0';
                for(; c<b && isspace(s[c]); ++c);
                if(s[c]==',')
                {
                    for(++c; c<b && isspace(s[c]); ++c);
                    for(prec=0; c<b && isdigit(s[c]); ++c)prec=prec*10+s[c]-'0';
                    for(; c<b && isspace(s[c]); ++c);
                }
                if(s[c]!=')')
                {
                    throw("Expected ')' in '" + s + "'");
                }
                ++c;
            }
            while(c<b)
            {
                string t;
                for(; c<b && isspace(s[c]); ++c);
                //fprintf(stderr, "Analyzing '%s'...\n", s.substr(c).c_str());
                if(c>=b)break;
                t = GetStr(s, c);
                if(t.empty())
                {
                    throw("Can't get a word from '" + s + "' at '" + s.substr(c) + "'");
                }
                if(t=="AUTO_INCREMENT")autoincrement = true;
                else if(t=="UNSIGNED")unsig = true;
                else if(t=="BINARY")binary = true;
                else if(t=="UNIQUE")unique = true;
                else if(t=="ZEROFILL")zerofill = true;
                else if(t=="FULLTEXT")fulltext = true;
                else if(t=="NULL")null = true;
                else if(t=="PRIMARY")
                {
                    for(; c<b && isspace(s[c]); ++c);
                    for(t=""; c<b && ischar(s[c]); ++c)t += toupper(s[c]);
                    if(t!="KEY")throw("PRIMARY "+t+"? KEY expected.");
                    /* ignore */
                }
                else if(t=="NOT")
                {
                    for(; c<b && isspace(s[c]); ++c);
                    for(t=""; c<b && ischar(s[c]); ++c)t += toupper(s[c]);
                    if(t!="NULL")throw("NOT "+t+"? NULL expected.");
                    null = false;
                }
                else if(t=="DEFAULT")
                {
                    for(; c<b && isspace(s[c]); ++c);
                    if(s[c]=='\'')
                    {
                        deftext = "";
                        while(c<b)
                        {
                            ++c;
                            if(s[c]=='\'')break;
                            if(s[c]=='\\')
                            {
                                ++c;
                                switch(s[c])
                                {
                                    case 'n': deftext += '\n'; break;
                                    case 'r': deftext += '\r'; break;
                                    case 't': deftext += '\t'; break;
                                    default: deftext += s[c];
                                }
                            }
                            else
                                deftext += s[c];
                        }
                        deftext = sqlfix(deftext);
                        ++c;
                    }
                    else
                    {
                        for(deftext=""; c<b && !isspace(s[c]); ++c)deftext += toupper(s[c]);
                        if(deftext=="NULL")
                            deftext="";
                        else
                            deftext = sqlfix(deftext);
                    }
                }
                else
                    throw("What is '" + t + "' of '" + s + "'? Type is '" + type + "'");
            }
            //fprintf(stderr, "(Done)\n");
        }
        if(type=="INT") { if(width==11)width=0; }
        else if(type=="BIGINT") { if(width==20)width=0; }
        else if(type=="TINYINT") { if(width==4)width=0; }
        else if(type=="SMALLINT") { if(width==6)width=0; }
        else if(type=="DOUBLE") { if(width==16 && prec==4)width=prec=0; }
        else if(type=="FLOAT") { if(width==10 && prec==2)width=prec=0; }
        else if(type=="CHAR") { if(width==1)width=0; }
        else if(type=="VARCHAR" && width<4)type="CHAR";
        
        RebuildContent();
    }
};
struct ForeignKeyData
{
    string ConstraintName;
    string IndexFields;
    string ReferenceTable;
    string ReferenceFields;
    string OnDelete;
    string OnUpdate;
    
    const std::string MakeString() const
    {
        std::string result;
        result = "(" + IndexFields + ")REFERENCES "+ReferenceTable
               + "(" + ReferenceFields + ")";
        if(!OnDelete.empty()) result += "ON DELETE " + OnDelete + " ";
        if(!OnUpdate.empty()) result += "ON UPDATE " + OnUpdate + " ";
        return result;
    }
    
    bool operator!= (const ForeignKeyData& b) const { return !(*this == b); }
    bool operator== (const ForeignKeyData& b) const
    {
        return IndexFields    == b.IndexFields
            && ReferenceTable == b.ReferenceTable
            && ReferenceFields == b.ReferenceFields
            && OnDelete == b.OnDelete
            && OnUpdate == b.OnUpdate;
    }
};

typedef pair<string, fieldtype> fieldpairtype;
typedef vector<fieldpairtype> fieldlisttype;
typedef pair<string, string> keypairtype;
typedef vector<keypairtype> keylisttype;
typedef pair<string, ForeignKeyData> foreignkeypairtype;
typedef vector<foreignkeypairtype> foreignkeylisttype;

string database = "winnie3";

class InputFailedError { public: bool pip; InputFailedError(bool p) : pip(p) { } };

static bool CaseEqual(const string &a, const string &b)
{
    unsigned aasi = a.size();
    unsigned beesi= b.size();
    if(aasi != beesi)return false;
    for(unsigned p=0; p<aasi; ++p)
        if(toupper(a[p]) != toupper(b[p]))
            return false;
    return true;
}

class Table
{
    string nam;
    string primarykey;
    
public:
    fieldlisttype fields;
    keylisttype keys;
    foreignkeylisttype fks;
    
#ifdef SUPPORT_INSERTS
    vector<string> inserts;
#endif
    
    Table(const string &name) : nam(name)
    {
        
    }
    void add(const string &name, const string &type)
    {
        //fprintf(stderr, "*** Table '%s': '%s' = '%s'\n", nam.c_str(), name.c_str(), type.c_str());
        fields.push_back(fieldpairtype(name, fieldtype(type)));
    }
    void addkey(const string &name, const string &type)
    {
        keys.push_back(keypairtype(name, type));
    }
    void addforeign(const string& indexcols, const ForeignKeyData& details)
    {
        fks.push_back(foreignkeypairtype(indexcols, details));
    }
    void setprimary(const string &n)
    {
        primarykey = n;
        for(;;)
        {
            unsigned a = primarykey.find(", ");
            if(a==primarykey.npos)break;
            primarykey.erase(a+1, 1);
        }
        // fprintf(stderr, "*** Table '%s': primary = '%s'\n", nam.c_str(), n.c_str());
    }
    inline const string &name() const { return nam; }
    inline const string &pkey() const { return primarykey; }
    bool hasfield(const string &n) const
    {
        unsigned a, b=fields.size();
        for(a=0; a<b; ++a)
            if(CaseEqual(fields[a].first, n))return true;
        return false;
    }
    const fieldpairtype &getfield(const string &n) const
    {
        unsigned a, b=fields.size();
        for(a=0; a<b; ++a)
            if(CaseEqual(fields[a].first, n))
                return fields[a];
        return fields[0]; // error
    }
    const string getkey(const string &search) const
    {
        unsigned a, b=keys.size();
        for(a=0; a<b; ++a)
            if(CaseEqual(keys[a].first, search))
                return keys[a].second;
        return "";
    }
    const ForeignKeyData getfk(const string &search) const
    {
        unsigned a, b=fks.size();
        for(a=0; a<b; ++a)
            if(CaseEqual(fks[a].first, search))
                return fks[a].second;
        return ForeignKeyData();
    }
#ifdef SUPPORT_INSERTS
    void Insert(const string &s)
    {
        inserts.push_back(s);
    }
    void FlushInserts() const
    {
        unsigned a, b=inserts.size();
        for(a=0; a<b; ++a)
            printf("insert into %s %s;\n", nam.c_str(), inserts[a].c_str());
    }
    inline bool tyhjataulu() const { return inserts.size() == 0; }
#endif
    void postfix()
    {
        unsigned a;
        bool hasvar = false;
        for(a=0; a<fields.size(); ++a)
        {
            if(fields[a].second.gettype()=="VARCHAR"
            || fields[a].second.gettype()=="TEXT"
            || fields[a].second.gettype()=="LONGBLOB"
            || fields[a].second.gettype()=="BLOB")
                hasvar = true;
        }
        if(hasvar)
        {
            for(a=0; a<fields.size(); ++a)
            {
                if(fields[a].second.gettype() == "CHAR"
                && fields[a].second.getwidth() > 3)
                    fields[a].second.settype("VARCHAR");
            }
        }
    }
};

typedef pair<string, Table> tabletype;
typedef map<string, Table> tablelisttype;

class sqldump
{
    tablelisttype tables;
    
    void load(const string &s) // analyzes the result
    {
        string Constraint;
        
        unsigned a, b=s.size();
        for(a=0; a<b; ++a)
        {
            if(s[a]=='#')
            {
                while(a+1<b && s[++a]!='\n');
                continue;
            }
            if((a+1) < b && s[a]=='-' && s[a+1]=='-')
            {
                while(a < b && s[a] != '\n')++a;
                continue;
            }
            #define isempty(x) ((x)=='\n' || (x)=='\r' || (x)=='\t' || (x)==' ')
            if(isempty(s[a]))continue;
            //fprintf(stderr, "*** Testing: '%s'\n", s.substr(a,50).c_str());
            if(startwith("CREATE ", s, a))
            {
                #define SKIPBLANK while(a < b && isempty(s[a]))++a;
                a += 7;
                SKIPBLANK
                if(startwith("INDEX ", s, a))
                {
                    throw string("Create index isn't working. Use key instead.");
                    a += 6;
                }
                if(!startwith("TABLE ", s, a))goto SkipCmd;
                a += 6;
                Table t(GetStr(s, a));
                SKIPBLANK
                if(s[a] == '(')++a;
                for(;;)
                {
                    SKIPBLANK
                    if(s[a]=='#')
                    {
                        while(a+1<b && s[++a]!='\n');
                        continue;
                    }
                    string fldname = GetStr(s, a);
                    
                    SKIPBLANK
                    if(ucase(fldname) == "PRIMARY")
                    {
                        if(ucase(GetStr(s, a)) != "KEY")
                        {
                            throw string("Primary what?");
                        }
                        
                        string content;
                        SKIPBLANK
                        if(s[a]=='(')++a;
                        while(a < b && s[a] != ')')content += s[a++];
                        ++a;
                        t.setprimary(content);
                    }
                    else if(ucase(fldname) == "KEY")
                    {
                        string k = GetStr(s, a);
                        string content;
                        SKIPBLANK
                        int sku=0;
                        while(a < b)
                        {
                            if(s[a] != ' ')content += s[a];
                            if(s[a] == '(')++sku;
                            else if(s[a] == ')')
                            {
                                if(!--sku)break;
                            }
                            ++a;
                        }
                        ++a;
                        t.addkey(k, content);
                    }
                    else if(ucase(fldname) == "UNIQUE"
                         || ucase(fldname) == "FULLTEXT")
                    {
                    RetryUnique:
                        string k = ucase(fldname) == "UNIQUE"
                                 ? "#"
                                 : "*";
                        k += GetStr(s, a);
                        if(ucase(k.substr(1)) == "KEY")
                        {
                            SKIPBLANK
                            goto RetryUnique;
                        }
                        
                        SKIPBLANK
                        int sku=0;
                        string content;
                        while(a < b)
                        {
                            SKIPBLANK
                            if(a >= b) break;
                            
                            content += s[a];
                            if(s[a] == '(')++sku;
                            if(s[a] == ')')
                            {
                                if(!--sku)break;
                            }
                            ++a;
                        }
                        if(a < b) ++a;
                        t.addkey(k, content);
                    }
                    else if(ucase(fldname) == "CONSTRAINT")
                    {
                        /* remember until the next FK definition */
                        Constraint = GetStr(s, a);
                    }
                    else if(ucase(fldname) == "FOREIGN")
                    {
                        ForeignKeyData fk;
                        fk.ConstraintName = Constraint;
                        
                        SKIPBLANK
                        if(ucase(GetStr(s, a)) != "KEY")
                        {
                            throw string("Foreign what?");
                        }

                        SKIPBLANK
                        if(s[a]=='(')
                            ++a;
                        else
                            throw string("Expected '(' after FOREIGN KEY");
                        while(a < b && s[a] != ')')
                        {
                            if(s[a] == ',') { fk.IndexFields += s[a++]; SKIPBLANK; continue; }
                            string part = GetStr(s, a);
                            if(part.empty())
                            {
                                throw string("Couldn't get FK part (got so far: "+fk.IndexFields+")");
                            }
                            fk.IndexFields += part;
                            SKIPBLANK
                        }
                        if(a < b && s[a] == ')')
                            ++a;
                        else
                            throw string("Expected ')' after FOREIGN KEY("+fk.IndexFields+")");
                        
                        SKIPBLANK
                        if(ucase(GetStr(s, a)) != "REFERENCES")
                        {
                            throw string("Expected: 'REFERENCES'");
                        }
                        
                        SKIPBLANK
                        fk.ReferenceTable = GetStr(s, a);
                        SKIPBLANK
                        if(s[a]=='(')
                            ++a;
                        else
                            throw string("Expected '(' after REFERENCES");
                        while(a < b && s[a] != ')')
                        {
                            if(s[a] == ',') { fk.ReferenceFields += s[a++]; SKIPBLANK; continue; }
                            string part = GetStr(s, a);
                            if(part.empty())
                            {
                                throw string("Couldn't get FK REF part ("+fk.IndexFields+")ref=("+fk.ReferenceFields+")");
                            }
                            fk.ReferenceFields += part;
                            SKIPBLANK
                        }
                        if(a < b && s[a] == ')')
                            ++a;
                        else
                            throw string("Expected ')' after REFERENCES("+fk.ReferenceFields+")");
                        
                        for(;;)
                        {
                            SKIPBLANK
                            if(startwith("ON DELETE", s, a))
                            {
                                a += 2+1+6; SKIPBLANK
                                string part = GetStr(s, a);
                                if(part == "NO" || part == "SET")
                                {
                                    part += " " + GetStr(s, a);
                                }
                                fk.OnDelete = ucase(part);
                                
                                if(fk.OnDelete == "RESTRICT") fk.OnDelete = "";                                
                                continue;
                            }
                            if(startwith("ON UPDATE", s, a))
                            {
                                a += 2+1+6; SKIPBLANK
                                string part = GetStr(s, a);
                                if(part == "NO" || part == "SET")
                                {
                                    part += " " + GetStr(s, a);
                                }
                                fk.OnUpdate = ucase(part);
                                continue;
                            }
                            break;
                        }
                        
                        if(fk.ConstraintName.empty())
                        {
                            static unsigned autonum = 1;
                            stringstream tmp;
                            tmp << "autogenerated_constraint_";
                            tmp << autonum++;
                            fk.ConstraintName = tmp.str();
                        }
                        
                        t.addforeign(fk.IndexFields, fk);
                        
                        Constraint = "";
                    }
                    else 
                    {
                        bool heitto = false;
                        bool empti = true;
                        int sulku = 0;
                        string content;
                        while(a < b && (sulku || (s[a] != ',' && s[a] != ')') || heitto))
                        {
                            if(s[a]=='#' && !heitto)
                            {
                                while(a+1<b && s[++a]!='\n');
                                continue;
                            }
                            if(isempty(s[a]))
                                empti = true;
                            else
                            {
                                if(empti && !content.empty())content += ' ';
                                content += heitto ? s[a] : toupper(s[a]);
                                empti = false;
                            }
                            if(s[a] == '(')++sulku;
                            else if(s[a] == ')')--sulku;
                            if(s[a] == '\\')content += s[++a];
                            else if(s[a] == '\'')heitto = !heitto;
                            ++a;
                        }
                        if(content.find("PRIMARY KEY") != content.npos)
                        {
                            t.setprimary(fldname);
                        }
                        t.add(fldname, content);
                    }
                    SKIPBLANK
                    if(s[a] == ')')break;
                    if(s[a] == ',')++a;
                }
                SKIPBLANK
                while(a+1 < b && s[a] != ';')++a;
                
                t.postfix();
                tables.insert(tabletype(t.name(), t));
                
                continue;
            }
#ifdef SUPPORT_INSERTS
            if(startwith("INSERT INTO ", s, a))
            {
                a += 11;
                SKIPBLANK;
                string tabname = GetStr(s, a);
                SKIPBLANK;
                bool heitto=false;
                string content;
                while(a < b && (s[a] != ';' || heitto))
                {
                    content += s[a];
                    if(s[a] == '\\')content += s[++a];
                    else if(s[a] == '\'')heitto = !heitto;
                    ++a;
                }
                if(!nodata)
                    tables.find(tabname)->second.Insert(content);
            }
#endif
SkipCmd:    bool heitto=false;
            while(a < b && (s[a] != ';' || heitto))
            {
                if(s[a] == '\\')++a;
                else if(s[a] == '\'')heitto = !heitto;
                ++a;
            }
        }
    }
public:
    sqldump(FILE *fp, bool ispipe=false)
    {
        if(!fp)
        {
            throw InputFailedError(ispipe);
        }
        fprintf(stderr, "*** Reading%s...\n", ispipe?" pipe":" file");
        char Buf[4096];
        string s;
        for(;;)
        {
            int n = fread(Buf, 1, sizeof Buf, fp);
            if(n <= 0)break;
            s += string(Buf, n);
        }
        if(ispipe)
            pclose(fp);
        else
            fclose(fp);
        fprintf(stderr, "*** Analyzing...\n");
        load(s);
    }
    
    
#if 1
    class alter_table_data
    {
    public:
        std::vector<std::string/*field name*/> DROP_FIELDS;
        std::vector<std::string/*field name*/> DROP_DEFAULTS;
        std::vector<std::string/*index name*/> DROP_INDEXES;
        std::vector<std::string/*fk name*/> DROP_FOREIGN_KEYS;
        bool DROP_PRIMARY_KEY;

        std::map<std::string/*field name*/, std::string/* field def*/> CHANGE_FIELDS;
        
        std::vector<std::pair<std::string/*field name*/, std::string/*field def*/> > ADD_FIELDS;
        std::vector<std::pair<std::string/*index name*/, std::string/*index def*/> > ADD_INDEXES;
        std::vector<std::string/*fk content*/> ADD_FOREIGN_KEYS;
        std::string ADD_PRIMARY;
    };
#endif
    class alter_db_data
    {
    public:
#if 1
        std::vector<std::string/*table name*/> DROP_TABLES;
        std::vector<std::string/*table content*/> CREATE_TABLES;
        std::map<std::string/*table name*/, alter_table_data> ALTER_TABLES;
#endif
        void Flush()
        {
            typedef std::vector<std::string>::const_iterator seti;
            typedef std::map<std::string, std::string>::const_iterator mapi;
            typedef std::vector<std::pair<std::string, std::string> >::const_iterator veci;
            
            /* First drop foreign keys to prevent innodb from complaining
             * when tables/keys are dropped */
            std::map<std::string, alter_table_data>::const_iterator i;
            for(i=ALTER_TABLES.begin(); i!=ALTER_TABLES.end(); ++i)
            {
                const std::string& tablename = i->first;
                const alter_table_data& data = i->second;

                std::vector<std::string> alter_list;

                for(seti j=data.DROP_FOREIGN_KEYS.begin(); j!=data.DROP_FOREIGN_KEYS.end(); ++j)
                    alter_list.push_back("DROP FOREIGN KEY " + *j);
                FlushAlterList(tablename, alter_list);
            }

            for(unsigned a=0; a<DROP_TABLES.size(); ++a)
                printf("DROP TABLE %s;\n", DROP_TABLES[a].c_str());
            
            for(unsigned a=0; a<CREATE_TABLES.size(); ++a)
                printf("CREATE TABLE %s;\n", CREATE_TABLES[a].c_str());
            
            for(i=ALTER_TABLES.begin(); i!=ALTER_TABLES.end(); ++i)
            {
                const std::string& tablename = i->first;
                const alter_table_data& data = i->second;
                
                printf("-- Table %s --\n", tablename.c_str());
                
                std::vector<std::string> alter_list;
                
                for(seti j=data.DROP_FIELDS.begin(); j!=data.DROP_FIELDS.end(); ++j)
                    alter_list.push_back("DROP " + *j);
                for(seti j=data.DROP_DEFAULTS.begin(); j!=data.DROP_DEFAULTS.end(); ++j)
                    alter_list.push_back("ALTER " + *j + " DROP DEFAULT");
                for(seti j=data.DROP_INDEXES.begin(); j!=data.DROP_INDEXES.end(); ++j)
                {
                    const char *mkey = j->c_str();
                    alter_list.push_back("DROP INDEX " + std::string
                              ((mkey[0]=='#' || mkey[0]=='*') ? mkey+1 : mkey));
                }
                if(data.DROP_PRIMARY_KEY)
                    alter_list.push_back("DROP PRIMARY KEY");
                
                FlushAlterList(tablename, alter_list);
                
                for(mapi j=data.CHANGE_FIELDS.begin(); j!=data.CHANGE_FIELDS.end(); ++j)
                    alter_list.push_back("CHANGE " + j->first + " " + j->second);
                    
                for(veci j=data.ADD_FIELDS.begin(); j!=data.ADD_FIELDS.end(); ++j)
                    alter_list.push_back("ADD " + j->first + " " + j->second);
                
                FlushAlterList(tablename, alter_list);
                
                for(veci j=data.ADD_INDEXES.begin(); j!=data.ADD_INDEXES.end(); ++j)
                {
                    const char *mkey = j->first.c_str();
                    alter_list.push_back("ADD " + std::string
                        (mkey[0]=='#' ? "UNIQUE"
                        :mkey[0]=='*' ? "FULLTEXT"
                        : "KEY")
                        + " "
                        + ((mkey[0]=='#' || mkey[0]=='*') ? mkey+1 : mkey)
                        + j->second);
                }
                
                if(!data.ADD_PRIMARY.empty())
                    alter_list.push_back("ADD PRIMARY KEY (" + data.ADD_PRIMARY + ")");

                FlushAlterList(tablename, alter_list);
            }
            
            /* Last add foreign keys. It must be done only after the keys are there. */
            for(i=ALTER_TABLES.begin(); i!=ALTER_TABLES.end(); ++i)
            {
                const std::string& tablename = i->first;
                const alter_table_data& data = i->second;
                
                std::vector<std::string> alter_list;
                
                for(seti j=data.ADD_FOREIGN_KEYS.begin(); j!=data.ADD_FOREIGN_KEYS.end(); ++j)
                {
                    alter_list.push_back("ADD FOREIGN KEY" + *j);
                }

                FlushAlterList(tablename, alter_list);
            }
        }
    private:
        void FlushAlterList(const std::string& tablename,
                            std::vector<std::string>& alter_list)
        {
            if(!alter_list.empty())
            {
                printf("ALTER TABLE %s ", tablename.c_str());
                for(unsigned a=0; a<alter_list.size(); ++a)
                {
                    if(a>0)printf(", ");
                    printf("%s", alter_list[a].c_str());
                }
                printf(";\n");
                alter_list.clear();
            }
        }
    };
    
    void updatewith(const sqldump &uusi)
    {
        fprintf(stderr, "*** Combining...\n");
        
        if(tables.empty())
            printf("CREATE DATABASE %s;\n", database.c_str());
        
        printf("USE %s;\n", database.c_str());
        
        alter_db_data changes;
        
        /* Find things that no longer apply */
        for(tablelisttype::const_iterator i=tables.begin(); i!=tables.end(); ++i)
        {
            const Table& oldtable = i->second;
            
            const std::string& tablename = i->first;
            
            tablelisttype::const_iterator j = uusi.tables.find(tablename);
            if(j == uusi.tables.end())
            {
                /* If the table no longer exists */
                changes.DROP_TABLES.push_back(tablename);
                continue;
            }
            
            const Table& newtable = j->second;
            
            changes.ALTER_TABLES[tablename].DROP_PRIMARY_KEY=false;
            
            fieldlisttype::const_iterator k;
            for(k = oldtable.fields.begin(); k != oldtable.fields.end(); ++k)
            {
                const string& fieldname = k->first;
                if(newtable.hasfield(fieldname))
                {
                    /* old field */
                    const fieldpairtype &oldfp = *k;
                    const fieldtype     &oldf  = oldfp.second;
                    const string olddefault = oldf.defval();
                    
                    /* new field */
                    const fieldpairtype &newfp = newtable.getfield(fieldname);
                    const fieldtype     &newf  = newfp.second;
                    const string newdefault = newf.defval();
                    
                    if(oldf != newf
                    || (newdefault != olddefault && !newdefault.empty()))
                    {
                        std::string def = newfp.first + " " + newf.c_str();
                        if(!newdefault.empty())
                        {
                            def += " DEFAULT " + newdefault;
                            if(newdefault != olddefault)
                                def += " -- Default was: " + olddefault + "\n";
                        }
                        def += string("\n -- Field was: ") + oldf.c_str() + "\n";
                        changes.ALTER_TABLES[tablename].CHANGE_FIELDS[fieldname] = def;
                    }
                    if(newdefault.empty() && !olddefault.empty())
                    {
                        std::string f = fieldname;
                        f += " -- Default was: " + olddefault + "\n";
                        changes.ALTER_TABLES[tablename].DROP_DEFAULTS.push_back(f);
                    }
                }
                else
                {
                    changes.ALTER_TABLES[tablename].DROP_FIELDS.push_back(fieldname);
                }
            }
            
            /* Drop extinct indices */
            keylisttype::const_iterator m;
            for(m = oldtable.keys.begin(); m != oldtable.keys.end(); ++m)
            {
                const string& keyname = m->first;
                const string newkey   = newtable.getkey(keyname);
                if(newkey.empty())
                {
                    std::string def = keyname + "\n -- Index was: " + m->second + "\n";
                    changes.ALTER_TABLES[tablename].DROP_INDEXES.push_back(def);
                }
            }
            
            /* Drop extinct foreign keys */
            foreignkeylisttype::const_iterator f;
            for(f = oldtable.fks.begin(); f != oldtable.fks.end(); ++f)
            {
                const string& keyname       = f->first;
                const ForeignKeyData newkey = newtable.getfk(keyname);
                if(newkey.IndexFields.empty())
                {
                    changes.ALTER_TABLES[tablename].DROP_FOREIGN_KEYS.push_back(newkey.ConstraintName);
                }
            }
        }
        
        /* Find all things that apply now */
        for(tablelisttype::const_iterator i=uusi.tables.begin(); i!=uusi.tables.end(); ++i)
        {
            const Table& newtable = i->second;
            
            bool create_table;
            std::string create_table_data;
            
            const std::string& tablename = i->first;
            
            tablelisttype::const_iterator j = tables.find(tablename);
            
            create_table = (j == tables.end());
            
            if(create_table)
            {
                /* If the table didn't exist before */
            }
            else
            {
                const Table& oldtable = j->second;
            
                if(oldtable.pkey() != newtable.pkey())
                {
                    /*Drop the primary key if it has changed */
                    changes.ALTER_TABLES[tablename].DROP_PRIMARY_KEY = true;
                }
            }
            
            bool primarydone = false;
            bool eka = true;
            string ed = "";
            
            /* Add new fields (drop & change already done) */
            fieldlisttype::const_iterator k;
            for(k = newtable.fields.begin(); k != newtable.fields.end(); ++k)
            {
                if(create_table)
                {
                    if(eka)eka=false;else create_table_data += ",\n";
                    
                    create_table_data += "  " + k->first + " " + k->second.c_str();
                    
                    const string defval = k->second.defval();
                    if(!defval.empty())
                        create_table_data += " DEFAULT " + defval;
                }
                else
                {
                    const Table& oldtable = j->second;
            
                    if(!oldtable.hasfield(k->first))
                    {
                        std::string def = k->second.c_str();
                        const string defval = k->second.defval();
                        if(!defval.empty())
                            def += " DEFAULT " + defval;

                        if(k->first == newtable.pkey())
                        {
                            primarydone = true;
                            def += " PRIMARY KEY ";
                        }
                        if(!ed.empty())
                            def += " AFTER " + ed;
                        else
                            def += " FIRST";

                        changes.ALTER_TABLES[tablename].ADD_FIELDS.
                            push_back(std::make_pair(k->first, def));
                    }
                }
                ed = k->first.c_str();
            }
            
            /* Update/add indices */
            keylisttype::const_iterator m;
            for(m = newtable.keys.begin(); m != newtable.keys.end(); ++m)
            {
                const string& keyname = m->first;
                const char *mkey = keyname.c_str();
                if(create_table)
                {
                    if(eka)eka=false;else create_table_data += ",\n";
                    
                    create_table_data += std::string
                        (mkey[0]=='#' ? "UNIQUE"
                        :mkey[0]=='*' ? "FULLTEXT"
                        : "KEY")
                        + " "
                        + ((mkey[0]=='#' || mkey[0]=='*') ? mkey+1 : mkey)
                        + m->second;
                }
                else
                {
                    const Table& oldtable = j->second;
            
                    const string oldkey = oldtable.getkey(keyname);
                    if(oldkey != m->second)
                    {
                        if(!oldkey.empty())
                        {
                            std::string def = keyname + "\n -- Index was: " + oldkey + "\n";
                            changes.ALTER_TABLES[tablename].DROP_INDEXES.push_back(def);
                        }
                        // m->second shouldn't be empty
                        changes.ALTER_TABLES[tablename].ADD_INDEXES.push_back(std::make_pair(keyname, m->second));
                    }
                }
            }
            
            /* Update/add foreign keys */
            foreignkeylisttype::const_iterator f;
            for(f = newtable.fks.begin(); f != newtable.fks.end(); ++f)
            {
                const string& keyname = f->first;
                
                std::string FKdef = f->second.MakeString();
                if(create_table)
                {
                    if(eka)eka=false;else create_table_data += ",\n";
                    create_table_data += std::string("FOREIGN KEY(") + keyname + ")" + FKdef;
                }
                else
                {
                    const Table& oldtable = j->second;
            
                    const ForeignKeyData oldkey = oldtable.getfk(keyname);
                    if(oldkey != f->second)
                    {
                        if(!oldkey.IndexFields.empty())
                        {
                            string def = oldkey.ConstraintName;
                            def += "\n -- FK Was: " + oldkey.MakeString() + "\n";
                            changes.ALTER_TABLES[tablename].DROP_FOREIGN_KEYS.push_back(def);
                        }
                        // f->second shouldn't be empty
                        changes.ALTER_TABLES[tablename].ADD_FOREIGN_KEYS.push_back(FKdef);
                    }
                }
            }
            
            if(create_table)
            {
                if(!newtable.pkey().empty())
                {
                    if(!eka) create_table_data += ",\n";
                    create_table_data += "  PRIMARY KEY(" + newtable.pkey() + ")";
                    eka = false;
                }
                if(!eka) create_table_data += "\n";
                changes.CREATE_TABLES.push_back(tablename + "(" + create_table_data + ")");
            }
            else
            {
                const Table& oldtable = j->second;
            
                if(!primarydone && oldtable.pkey() != newtable.pkey())
                {
                    changes.ALTER_TABLES[j->first].ADD_PRIMARY = newtable.pkey();
                }
            }
            
#ifdef SUPPORT_INSERTS
            if(create_table || oldtable.tyhjataulu())
            {
                newtable.FlushInserts();
            }
#endif
        }
        changes.Flush();
    }
};

static void Usage(void)
{
    printf(
        "sqlupdate v"VERSION" - Copyright (C) Joel Yliluoma (http://iki.fi/bisqwit/)\n"
        "\n"
        "Usage:\n"
        "    sqlupdate [options] >changes.sql\n"
        "       (Creates an update script)\n"
        "\n"
        "Options:\n"
        "    -t tablefile          Describes the file containing\n"
        "                          the new sql layout. Default: tables.sql\n"
        "    -d database           Default: winnie3\n"
        "    -h host               Default: localhost\n"
        "    -u user               Default: root\n"
        "    -p pass\n"
        "    -r                    Reverse operation. (new->old)\n"
#ifdef SUPPORT_INSERTS
        "    -D                    Insert new data\n"
#endif
        "\n"
        "Example:\n"
        "  ./sqlupdate -t etimgr.sql -d etimgr | mysql -uroot\n"
        "\n"
        "This program does not update a database. It only\n"
        "produces update scripts (which show the differences).\n\n");
}

int main(int argc, const char *const *argv)
{
    string uusifile = "tables.sql";
    string host = "localhost";
    string user = "root";
    string pass = ""; 
    int vanhaksi = 0;
    
    while(--argc > 0)
    {
        const char *s = *++argv;
        if(*s=='-')
            while(*++s)
                switch(*s)
                {
                    #define stringparm(svar) \
                        if(*++s){svar = s;s=" ";}else{if(!--argc){Usage();return -1;};svar=*++argv;s=" ";}
                    
                    case 't': stringparm(uusifile); break;
                    case 'd': stringparm(database); break;
                    case 'h': stringparm(host); break;
                    case 'u': stringparm(user); break;
                    case 'p': stringparm(pass); break;
#ifdef SUPPORT_INSERTS
                    case 'D':
                        nodata = false;
                        break;
#endif
                    case 'r':
                        vanhaksi = 1;
                        break;
                    default:
                        Usage();
                        return -1;
                }
        else
        {
            Usage();
            return -1;
        }
    }
    if(vanhaksi < 0) { Usage(); return -1; }
        
    string vanhakomento = "/usr/bin/mysqldump -e -u'" + user + "' -h'" + host + "'";
#ifdef SUPPORT_INSERTS
    if(nodata)
#endif
        vanhakomento += " -d";
    
    if(!pass.empty())
        vanhakomento += " -p'" + pass + "'";
    
    vanhakomento += " '";
    vanhakomento += database;
    vanhakomento += "'";
    
    try
    {
        sqldump vanha(popen(vanhakomento.c_str(), "r"), true);
        sqldump uusi(fopen(uusifile.c_str(), "r"));
    
        if(vanhaksi==1)
            uusi.updatewith(vanha);
        else
            vanha.updatewith(uusi);
    }
    catch(const InputFailedError &p)
    {
        fprintf(stderr, "*** %s failed!\n", p.pip ? database.c_str() : uusifile.c_str());
    }
    catch(const string &s)
    {
        fprintf(stderr, "*** Oops.\n*** %s\n", s.c_str());
    }
        
    fprintf(stderr, "*** Done\n");
    
    return 0;
}
